import java.io.*;
import java.util.*;

public class ImportShpMap { 
    static final int SHP_MAGIC = 9994;
    static final int DBF_MAGIC = 0x6603;

    // Shape types
    static final int NULL       = 0;
    static final int POINT      = 1;
    static final int POLY_LINE  = 3;
    static final int POLYGON    = 5;
    static final int MULTIPOINT = 8;

    static final int SPEED[] = {60, 40, 20}; // average speed km/h for streets at different levels

    // Type mapping
    static final int dbfTypeMap[] = { ObjectType.BANK, ObjectType.WATER, ObjectType.MONUMENT, ObjectType.HOTEL, ObjectType.THEATRE, ObjectType.POLICE }; 



    static double readDoubleLE(byte b[], int p) { 
        return Double.longBitsToDouble(readLongLE(b, p));
    }

    static long readLongLE(byte b[], int p) { 
	return ((long)readIntLE(b, p+4) << 32) | (readIntLE(b, p) & 0xFFFFFFFFL);
    }

    static int readIntLE(byte b[], int p) { 
	return (b[p+3] << 24) | ((b[p+2] & 0xFF) << 16) | ((b[p+1] & 0xFF) << 8) | (b[p] & 0xFF);
    }
    static int readWordLE(byte b[], int p) { 
	return ((b[p+1] & 0xFF) << 8) | (b[p] & 0xFF);
    }

    static long readLongBE(byte b[], int p) { 
	return (long)(readIntBE(b, p) << 32) | (readIntBE(b, p+4) & 0xFFFFFFFFL);
    }

    static int readIntBE(byte b[], int p) { 
	return (b[p] << 24) | ((b[p+1] & 0xFF) << 16) | ((b[p+2] & 0xFF) << 8) | (b[p+3] & 0xFF);
    }
    static int readWordBE(byte b[], int p) { 
	return ((b[p] & 0xFF) << 8) | (b[p+1] & 0xFF);
    }

    static class DbfField { 
        char   type;
        int    size;
        int    prec;
        String name;
    }

    static class DbfRecord { 
        byte[]     data;
        DbfField[] field;
        
        DbfRecord(int size, DbfField[] field) { 
            data = new byte[size];
            this.field = field;
        }

        HashMap read(InputStream in) throws IOException { 
            in.read(data);
	    if (data[0] == '*') { 		
                return null;
            }
            HashMap map = new HashMap();
            
            int len, tmp;
            String str;
            for (int i = 0, k = 1; i < field.length; i++) { 
                Object val = null;
                switch (field[i].type) { 
                case 'C':
                    len = field[i].size + ((field[i].prec & 0xFF) << 8);
                    val = new String(data, k, len).trim();
                    k += len;
                    break;
                case 'N':
                    str = new String(data, k, field[i].size).trim();
                    if (str.length() == 0) { 
                        str = "0";
                    }
                    k += field[i].size;
                    val = str;
                    break;
                case 'L':
                    tmp = data[k++] & ~('a' - 'A');
                    val = Boolean.valueOf(tmp == 'Y' || tmp == 'T');
                    break;
                case 'I':
                    val = new Integer(readIntLE(data, k));
                    k += 4;
                    break;
                case 'D':	
                    str = new String(data, k, 8).trim();
                    if (str.length() != 0) { 
                        val = java.sql.Date.valueOf(str.substring(0, 4) + "-" + 
                                                    str.substring(4, 6) + "-" +
                                                    str.substring(6, 8));
                    } else { 
                        val = null;
                    }
                    k += 8;
                    break;
                case 'M':
                    k += 10;
                    break;
                default:
                    k += field[i].size;
                }
                map.put(field[i].name, val);
            }
            return map;
        }
    }
            
    

    public static void main(String[] args) throws Exception { 
        TreeMap streets = new TreeMap();

        int nStreets = 0;

        for (int n = 0; n < args.length; n++) { 
            InputStream shp = new FileInputStream(args[n] + ".shp");
            InputStream dbf = new FileInputStream(args[n] + ".dbf");
            byte[] shpHdr = new byte[100];
            byte[] dbfHdr = new byte[32];
            if (shp.read(shpHdr) != shpHdr.length) { 
                throw new Error("Failed to read SHP header");
            }
            if (dbf.read(dbfHdr) != dbfHdr.length) { 
                 throw new Error("Failed to read DBF header");
            }
            if (readIntBE(shpHdr, 0) != SHP_MAGIC) { 
                throw new Error("Invalid shape file format");
            }
            if (readWordLE(dbfHdr, 0) != DBF_MAGIC) { 
                throw new Error("Invalid dbf file format");
            }
            int shpContentLen = readIntBE(shpHdr, 24)*2 - shpHdr.length;
            
            int dbfRecordNumber = readIntLE(dbfHdr, 4);
            int dbfHdrLen = readWordLE(dbfHdr, 8);
            int dbfRecLen = readWordLE(dbfHdr, 10);
            int dbfFieldNumber = (dbfHdrLen - 32) >>> 5;
            byte dbfFieldsData[] = new byte[dbfHdrLen - dbfHdr.length];

            if (dbf.read(dbfFieldsData) != dbfFieldsData.length) { 
                throw new Error("Failed to read DBF field info");
            }
              
            DbfField[] dbfField = new DbfField[dbfFieldNumber];
            for (int i = 0, j = 0; i < dbfFieldNumber; i += 1, j += 32) { 
                dbfField[i] = new DbfField();
                dbfField[i].name = new String(dbfFieldsData, j, 10).trim(); 
                dbfField[i].size = dbfFieldsData[j+16] & 0xFF;
                dbfField[i].prec = dbfFieldsData[j+17];
                dbfField[i].type = (char)((dbfFieldsData[j+11] & 0xFF) & ~('a' - 'A'));
            }                

            DbfRecord dbfRecord = new DbfRecord(dbfRecLen, dbfField);
            byte shpData[] = new byte[shpContentLen];
            
            if (shp.read(shpData) != shpData.length) { 
                throw new Error("Failed to read shape file data");
            }

            int counter = 0;
            for (int shpOffs = 0; --dbfRecordNumber >= 0;) { 
                HashMap columns = dbfRecord.read(dbf);
                if (columns != null) { 
                    int streetLevel = 0;
                    Object dbfType = columns.get("TYP");
                    String label = (String)columns.get("NAME");
                    int type = dbfType != null 
                        ? dbfTypeMap[Integer.parseInt((String)dbfType)] : ObjectType.STREET;
                    Street street = new Street(label, streetLevel, type);
                    nStreets += 1;
                    street.collision = (Street)streets.get(label);
                    streets.put(label, street);  
                
                    int shpRecordNo = readIntBE(shpData, shpOffs);
                    shpOffs += 4;
                    int shpRecordSize = readIntBE(shpData, shpOffs);
                    shpOffs += 4;
                    int shpType = readIntLE(shpData, shpOffs);

                    System.out.println("Name='" + label + "' type=" + ObjectType.name[type] + ", shape type=" + shpType);
                    
                    switch (shpType) { 
                    case POINT:
                    { 
                        double x = readDoubleLE(shpData, shpOffs+4);
                        double y = readDoubleLE(shpData, shpOffs+12);
                        Segment s = new Segment(street);
                        street.segments.add(s);
                        s.points.add(new Point(x, y, s));
                        break;
                    }
                    case MULTIPOINT:
                    { 
                        System.err.println("Leave only one point from multipoints");
                        double x = readDoubleLE(shpData, shpOffs+40);
                        double y = readDoubleLE(shpData, shpOffs+48);
                        Segment s = new Segment(street);
                        street.segments.add(s);
                        s.points.add(new Point(x, y, s));
                        break;
                    }                        
                    case POLY_LINE:
                    {
                        int nParts = readIntLE(shpData, shpOffs+36);
                        int nPoints = readIntLE(shpData, shpOffs+40);
                        int[] parts = new int[nParts+1];
                        for (int i = 0; i < nParts; i++) { 
                            parts[i] = readIntLE(shpData, shpOffs+44+i*4);
                        }
                        parts[nParts] = nPoints;
                        int offs = shpOffs+44+4*nParts;
                        for (int i = 0; i < nParts; i++) {                             
                            Segment s = new Segment(street);
                            street.segments.add(s);
                            for (int j = parts[i]; j < parts[i+1]; j++) { 
                                double x = readDoubleLE(shpData, offs);
                                double y = readDoubleLE(shpData, offs+8);
                                offs += 16;
                                s.points.add(new Point(x, y, s));
                            }
                        }
                        break;
                    }                               
                    case POLYGON:
                    {
                        int nParts = readIntLE(shpData, shpOffs+36);
                        int nPoints = readIntLE(shpData, shpOffs+40);
                        int[] parts = new int[nParts+1];
                        for (int i = 0; i < nParts; i++) { 
                            parts[i] = readIntLE(shpData, shpOffs+44+i*4);
                        }
                        parts[nParts] = nPoints;
                        int offs = shpOffs+44+4*nParts;
                        if (nParts > 1) { 
                            System.err.println("Only one segment of polygon is left");
                            nParts = 1;
                        }
                        for (int i = 0; i < nParts; i++) {                             
                            Segment s = new Segment(street);
                            street.segments.add(s);
                            for (int j = parts[i]; j < parts[i+1]; j++) { 
                                double x = readDoubleLE(shpData, offs);
                                double y = readDoubleLE(shpData, offs+8);
                                offs += 16;
                                s.points.add(new Point(x, y, s));
                            }
                            s.points.add(s.points.get(0));
                        }
                        break;
                    }                               
                    default:
                        System.err.println("Unsupported shape type " + type);
                    }
                    shpOffs += shpRecordSize*2;
                }                
            }
            dbf.close();
            shp.close();
        }    
        MapWriter.write(nStreets, streets, new MetricGrid(), SPEED, true, false);
    }
}